/*  demolition - destruction test LADSPA plugins
    Copyright (C) 2002,2003  Nick Lamb <njl195@zepler.org.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#define _GNU_SOURCE

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include <float.h>
#include <dlfcn.h>
#include <dirent.h>
#include <sys/types.h>
#include <signal.h>
#include <sys/wait.h>
#include <unistd.h>
#include <glib.h>

#include "ladspa.h"

FILE *log_file;

#ifndef PATH_MAX
#define PATH_MAX 4096
#endif

/* 1024 samples per block, unless there's some reason to do otherwise */
#define SAMPLES 1024
/* assert 48kHz sample rate, unless there's some reason to do otherwise */
#define SRATE 48000

/* arbitrary bits which are unlikely to occur by accident in floats */
/* should NEVER be READ or OVERWRITTEN */
#define DEADBEEF 0xdeadbeef
/* should ALWAYS be OVERWRITTEN */
#define FACEACHE 0xfaceacce

typedef struct {
  char *filename;
  long int index;
} ladspa_plugin;

typedef struct {
  void *library;
  const LADSPA_Descriptor *descriptor;
} plugin_code;

static void test (ladspa_plugin * plugin);
static void check_buffers(const LADSPA_Descriptor *descriptor,
                          LADSPA_Data *buffers[],
                          int size, int do_input);
static void setup_buffers(LADSPA_Handle *handle,
                         const LADSPA_Descriptor *descriptor,
                         LADSPA_Data *buffers[], int size);
static void setup_ports(LADSPA_Handle *handle,
                        const LADSPA_Descriptor *descriptor,
                        LADSPA_Data *controls);
static LADSPA_Handle instantiate(const LADSPA_Descriptor *descriptor,
                                 unsigned long rate);
static plugin_code * load (char *filename, long int index);
static void unload (plugin_code *code);

void log_comment(const char *format, ...);
void log_error(const char *format, ...);
static void find_all_plugins(void);
static void find_plugins(char *path_entry);
static void lib_plugins(char *lib_name);

static GSList *plugin_list;

static void test_buffer_junk (const LADSPA_Descriptor *descriptor)
{
  LADSPA_Handle *handle;
  LADSPA_Data *controls;
  LADSPA_Data **buffers;

  controls= calloc(descriptor->PortCount, sizeof(LADSPA_Data));
  buffers= calloc(descriptor->PortCount, sizeof(LADSPA_Data *));

  log_comment("# Running... test_buffer_junk\n");
  if ((handle = instantiate(descriptor, SRATE))) {
    unsigned long port;

    if (descriptor->activate) descriptor->activate(handle);
    setup_ports(handle, descriptor, controls);
    setup_buffers(handle, descriptor, buffers, SAMPLES);
    
    log_comment("# quiet phase, all input samples 0.0\n");
    descriptor->run(handle, SAMPLES);
    for (port= 0; port < descriptor->PortCount; ++port) {
      if (LADSPA_IS_PORT_AUDIO(descriptor->PortDescriptors[port])) {
        LADSPA_Data *buffer= buffers[port];

        if (LADSPA_IS_PORT_INPUT(descriptor->PortDescriptors[port])) {
          int k;
          for (k= 0; k<SAMPLES; ++k)
            buffer[k+1]= -1.0f;
        }
      }
    }
    log_comment("# all input samples -1.0\n");
    descriptor->run(handle, SAMPLES);
    check_buffers(descriptor, buffers, SAMPLES, 0);

    setup_buffers(handle, descriptor, buffers, SAMPLES);
    for (port= 0; port < descriptor->PortCount; ++port) {
      if (LADSPA_IS_PORT_AUDIO(descriptor->PortDescriptors[port])) {
        LADSPA_Data *buffer= buffers[port];

        if (LADSPA_IS_PORT_INPUT(descriptor->PortDescriptors[port])) {
          int k;
          for (k= 0; k<SAMPLES; ++k)
            buffer[k+1]= 2.0f;
        }
      }
    }
    log_comment("# all input samples 2.0\n");
    descriptor->run(handle, SAMPLES);
    check_buffers(descriptor, buffers, SAMPLES, 0);

    setup_buffers(handle, descriptor, buffers, SAMPLES);
    for (port= 0; port < descriptor->PortCount; ++port) {
      if (LADSPA_IS_PORT_AUDIO(descriptor->PortDescriptors[port])) {
        LADSPA_Data *buffer= buffers[port];

        if (LADSPA_IS_PORT_INPUT(descriptor->PortDescriptors[port])) {
          int k;
          for (k= 0; k<SAMPLES; ++k)
            buffer[k+1]= INFINITY;
        }
      }
    }
    log_comment("# all input samples +Inf\n");
    descriptor->run(handle, SAMPLES);

    if (descriptor->deactivate) descriptor->deactivate(handle);
    check_buffers(descriptor, buffers, SAMPLES, 0);
    descriptor->cleanup(handle);
  }

  free(buffers);
  free(controls);
}

static void test_control_extreme (const LADSPA_Descriptor *descriptor)
{
  LADSPA_Handle *handle;
  LADSPA_Data *controls;
  LADSPA_Data **buffers;

  controls= calloc(descriptor->PortCount, sizeof(LADSPA_Data));
  buffers= calloc(descriptor->PortCount, sizeof(LADSPA_Data *));

  log_comment("# Running... test_control_extreme\n");
  if ((handle = instantiate(descriptor, SRATE))) {
    unsigned long port;

    if (descriptor->activate) descriptor->activate(handle);
    setup_ports(handle, descriptor, controls);
    setup_buffers(handle, descriptor, buffers, SAMPLES);
    
    log_comment("# quiet phase\n");
    descriptor->run(handle, SAMPLES);
    for (port= 0; port < descriptor->PortCount; ++port) {
      if (LADSPA_IS_PORT_CONTROL(descriptor->PortDescriptors[port])
       && LADSPA_IS_PORT_INPUT(descriptor->PortDescriptors[port])) {
        controls[port]= 0.0f;
      }
    }
    log_comment("# all input controls zero\n");
    descriptor->run(handle, SAMPLES);
    for (port= 0; port < descriptor->PortCount; ++port) {
      if (LADSPA_IS_PORT_CONTROL(descriptor->PortDescriptors[port])
       && LADSPA_IS_PORT_INPUT(descriptor->PortDescriptors[port])) {
        controls[port]= FLT_MAX;
      }
    }
    log_comment("# all input controls FLOAT MAX\n");
    descriptor->run(handle, SAMPLES);
    for (port= 0; port < descriptor->PortCount; ++port) {
      if (LADSPA_IS_PORT_CONTROL(descriptor->PortDescriptors[port])
       && LADSPA_IS_PORT_INPUT(descriptor->PortDescriptors[port])) {
        controls[port]= FLT_MIN;
      }
    }
    log_comment("# all input controls FLOAT MIN\n");
    descriptor->run(handle, SAMPLES);
    for (port= 0; port < descriptor->PortCount; ++port) {
      if (LADSPA_IS_PORT_CONTROL(descriptor->PortDescriptors[port])
       && LADSPA_IS_PORT_INPUT(descriptor->PortDescriptors[port])) {
        controls[port]= INFINITY;
      }
    }
    log_comment("# all input controls +Inf\n");
    descriptor->run(handle, SAMPLES);
    for (port= 0; port < descriptor->PortCount; ++port) {
      if (LADSPA_IS_PORT_CONTROL(descriptor->PortDescriptors[port])
       && LADSPA_IS_PORT_INPUT(descriptor->PortDescriptors[port])) {
        controls[port]= -INFINITY;
      }
    }
    log_comment("# all input controls -Inf\n");
    descriptor->run(handle, SAMPLES);

    if (descriptor->deactivate) descriptor->deactivate(handle);
    check_buffers(descriptor, buffers, SAMPLES, 1);
    descriptor->cleanup(handle);
  }

  free(buffers);
  free(controls);
}

static void test_controls (const LADSPA_Descriptor *descriptor)
{
  LADSPA_Handle *handle;
  LADSPA_Data *controls;
  LADSPA_Data **buffers;

  controls= calloc(descriptor->PortCount, sizeof(LADSPA_Data));
  buffers= calloc(descriptor->PortCount, sizeof(LADSPA_Data *));

  log_comment("# Running... test_controls\n");
  if ((handle = instantiate(descriptor, SRATE))) {
    unsigned long port;

    if (descriptor->activate) descriptor->activate(handle);
    setup_ports(handle, descriptor, controls);
    setup_buffers(handle, descriptor, buffers, SAMPLES);
    
    log_comment("# quiet phase\n");
    descriptor->run(handle, SAMPLES);
    for (port= 0; port < descriptor->PortCount; ++port) {
      if (LADSPA_IS_PORT_CONTROL(descriptor->PortDescriptors[port])
       && LADSPA_IS_PORT_INPUT(descriptor->PortDescriptors[port])) {

        if (LADSPA_IS_HINT_BOUNDED_BELOW(
            descriptor->PortRangeHints[port].HintDescriptor))
        controls[port]= descriptor->PortRangeHints[port].LowerBound;
      }
    }
    log_comment("# all input controls lower bound (or default)\n");
    descriptor->run(handle, SAMPLES);

    setup_ports(handle, descriptor, controls);
    for (port= 0; port < descriptor->PortCount; ++port) {
      if (LADSPA_IS_PORT_CONTROL(descriptor->PortDescriptors[port])
       && LADSPA_IS_PORT_INPUT(descriptor->PortDescriptors[port])) {

        if (LADSPA_IS_HINT_BOUNDED_ABOVE(
            descriptor->PortRangeHints[port].HintDescriptor))
        controls[port]= descriptor->PortRangeHints[port].UpperBound;
      }
    }
    log_comment("# all input controls upper bound (or default)\n");
    descriptor->run(handle, SAMPLES);

    if (descriptor->deactivate) descriptor->deactivate(handle);
    check_buffers(descriptor, buffers, SAMPLES, 1);
    descriptor->cleanup(handle);
  }

  free(buffers);
  free(controls);
}

static void test_buffer (const LADSPA_Descriptor *descriptor)
{
  LADSPA_Handle *handle;
  LADSPA_Data *controls;
  LADSPA_Data **buffers;

  controls= calloc(descriptor->PortCount, sizeof(LADSPA_Data));
  buffers= calloc(descriptor->PortCount, sizeof(LADSPA_Data *));

  log_comment("# Running... test_buffer\n");
  if ((handle = instantiate(descriptor, SRATE))) {

    setup_ports(handle, descriptor, controls);
    if (descriptor->activate) descriptor->activate(handle);

    log_comment("# 4096 samples\n");
    setup_buffers(handle, descriptor, buffers, 4096);
    descriptor->run(handle, 4096);
    check_buffers(descriptor, buffers, 4096, 1);

    log_comment("# 16 samples\n");
    setup_buffers(handle, descriptor, buffers, 16);
    descriptor->run(handle, 16);
    check_buffers(descriptor, buffers, 16, 1);

    log_comment("# 1 sample\n");
    setup_buffers(handle, descriptor, buffers, 1);
    descriptor->run(handle, 1);
    check_buffers(descriptor, buffers, 1, 1);

    if (descriptor->deactivate) descriptor->deactivate(handle);
    descriptor->cleanup(handle);
  }

  free(buffers);
  free(controls);
}

static void test_lifecycle (const LADSPA_Descriptor *descriptor)
{
  LADSPA_Handle *handle;
  LADSPA_Data *controls;
  LADSPA_Data **buffers;

  controls= calloc(descriptor->PortCount, sizeof(LADSPA_Data));
  buffers= calloc(descriptor->PortCount, sizeof(LADSPA_Data *));

  log_comment("# Running... test_lifecycle\n");

  log_comment("# instantiate\n");
  if ((handle = instantiate(descriptor, SRATE))) {
    descriptor->cleanup(handle);
  }

  log_comment("# activate/deactivate\n");
  if ((handle = instantiate(descriptor, SRATE))) {
    if (descriptor->activate) descriptor->activate(handle);
    if (descriptor->deactivate) descriptor->deactivate(handle);
    descriptor->cleanup(handle);
  }

  log_comment("# run once, one sample\n");
  if ((handle = instantiate(descriptor, SRATE))) {
    if (descriptor->activate) descriptor->activate(handle);
    setup_ports(handle, descriptor, controls);
    setup_buffers(handle, descriptor, buffers, 1);
    descriptor->run(handle, 1);
    if (descriptor->deactivate) descriptor->deactivate(handle);
    check_buffers(descriptor, buffers, 1, 1);

    /* Should still be OK here */
    setup_ports(handle, descriptor, controls);

    if (descriptor->run_adding) {
      log_comment("# run adding, one sample\n");
      if (descriptor->activate) descriptor->activate(handle);
      descriptor->run_adding(handle, 1);
      descriptor->set_run_adding_gain(handle, 0.0f);
      if (descriptor->deactivate) descriptor->deactivate(handle);
    }
    descriptor->cleanup(handle);
  }

  free(buffers);
  free(controls);
}

static void test_description (const LADSPA_Descriptor *descriptor)
{
  unsigned long port;
  size_t k;

  log_comment("# UniqueID = %ld\n", descriptor->UniqueID);
  if (descriptor->UniqueID > 0x1000000) {
    log_error("UniqueID outside range from LADSPA 1.1\n");
  } else if (descriptor->UniqueID < 1000) {
    log_error("UniqueID from reserved range DO NOT DISTRIBUTE\n");
  }
  log_comment("# Label \"%s\"\n", descriptor->Label);
  if (descriptor->Label == NULL) {
    log_error("Label is illegal NULL string\n");
  }
  for (k = 0; k < strlen(descriptor->Label); ++k) {
    if (isspace(descriptor->Label[k])) {
      log_error("Label contains illegal white-space character\n");
      break;
    }
  }

  if (LADSPA_IS_REALTIME(descriptor->Properties)) {
    log_comment("# is real-time dependent\n");
  }
  if (LADSPA_IS_INPLACE_BROKEN(descriptor->Properties)) {
    log_comment("# is broken for in-place processing\n");
  }
  if (LADSPA_IS_HARD_RT_CAPABLE(descriptor->Properties)) {
    log_comment("# is suitable for hard real-time use\n");
  }
  if (descriptor->Properties > 0x7) {
    log_comment("# has unidentified additional properties\n");
  }

  log_comment("# Name \"%s\"\n", descriptor->Name);
  if (descriptor->Name == NULL) {
    log_error("Name is illegal NULL string\n");
  }
  log_comment("# Maker \"%s\"\n", descriptor->Maker);
  if (descriptor->Maker == NULL) {
    log_error("Maker is illegal NULL string\n");
  }
  log_comment("# Copyright \"%s\"\n", descriptor->Copyright);
  if (descriptor->Copyright == NULL) {
    log_error("Copyright is illegal NULL string\n");
  }
  log_comment("# PortCount = %lu\n", descriptor->PortCount);

  for (port = 0; port < descriptor->PortCount; ++port) {
    LADSPA_PortDescriptor about= descriptor->PortDescriptors[port];
    LADSPA_PortRangeHint hint = descriptor->PortRangeHints[port];

    log_comment("# Port %lu \"%s\"\n", port, descriptor->PortNames[port]);
    if (LADSPA_IS_PORT_INPUT(about) && LADSPA_IS_PORT_OUTPUT(about)) {
      log_error("port %lu is both INPUT and OUTPUT\n", port);
    } else if (!LADSPA_IS_PORT_INPUT(about) && !LADSPA_IS_PORT_OUTPUT(about)) {
      log_error("port %lu is neither INPUT nor OUTPUT\n", port);
    }
    if (LADSPA_IS_PORT_CONTROL(about) && LADSPA_IS_PORT_AUDIO(about)) {
      log_error("port %lu is both CONTROL and AUDIO\n", port);
    } else if (!LADSPA_IS_PORT_CONTROL(about) && !LADSPA_IS_PORT_AUDIO(about)) {
      log_error("port %lu is neither CONTROL nor AUDIO\n", port);
    }

    if (LADSPA_IS_HINT_BOUNDED_BELOW(hint.HintDescriptor)
     && LADSPA_IS_HINT_BOUNDED_ABOVE(hint.HintDescriptor)
     && hint.LowerBound > hint.UpperBound) {
      log_error("port %lu has LowerBound larger than UpperBound\n", port);
    }

    switch (hint.HintDescriptor & LADSPA_HINT_DEFAULT_MASK) {
      case LADSPA_HINT_DEFAULT_MINIMUM:
        if (!LADSPA_IS_HINT_BOUNDED_BELOW(hint.HintDescriptor))
          log_error("port %lu has DEFAULT_MINIMUM without BOUNDED_BELOW\n", port);
        break;
   
      case LADSPA_HINT_DEFAULT_LOW:
      case LADSPA_HINT_DEFAULT_MIDDLE:
      case LADSPA_HINT_DEFAULT_HIGH:
        if (!LADSPA_IS_HINT_BOUNDED_BELOW(hint.HintDescriptor))
          log_error("port %lu has DEFAULT without BOUNDED_BELOW\n", port);
        if (!LADSPA_IS_HINT_BOUNDED_ABOVE(hint.HintDescriptor))
          log_error("port %lu has DEFAULT without BOUNDED_ABOVE\n", port);
        break;
  
      case LADSPA_HINT_DEFAULT_MAXIMUM:
        if (!LADSPA_IS_HINT_BOUNDED_ABOVE(hint.HintDescriptor))
          log_error("port %lu has DEFAULT_MAXIMUM without BOUNDED_ABOVE\n", port);
        break;

      default:
        break;
    }

    if (LADSPA_IS_HINT_TOGGLED(hint.HintDescriptor)) {
      LADSPA_PortRangeHintDescriptor prhd;
      prhd = hint.HintDescriptor;
      if (prhd & 0x3b) {
        log_error("port %lu uses illegal combination of hints with TOGGLED\n", port);
      }
      if (LADSPA_IS_HINT_HAS_DEFAULT(prhd) && !LADSPA_IS_HINT_DEFAULT_0(prhd) && !LADSPA_IS_HINT_DEFAULT_1(prhd)) {
        log_error("port %lu uses illegal default with TOGGLED\n", port);
      }
    }

    if (LADSPA_IS_HINT_LOGARITHMIC(hint.HintDescriptor)) {
      LADSPA_PortRangeHintDescriptor prhd;
      prhd = hint.HintDescriptor;
      if (LADSPA_IS_HINT_BOUNDED_BELOW(prhd) && hint.LowerBound <= 0.0f) {
        log_error("port %lu is LOGARITHMIC but LowerBound isn't positive\n", port);
      }
      if (LADSPA_IS_HINT_BOUNDED_ABOVE(prhd) && hint.UpperBound <= 0.0f) {
        log_error("port %lu is LOGARITHMIC but UpperBound isn't positive\n", port);
      }
    }
  }

  if (descriptor->activate) {
      log_comment("# activate() is needed\n");
  }
  if (descriptor->run_adding) {
    if (descriptor->set_run_adding_gain) {
      log_comment("# run_adding() is implemented\n");
    } else {
      log_error("run_adding() implemented without set_run_adding_gain()\n");
    }
  } else {
    if (descriptor->set_run_adding_gain) {
      log_error("set_run_adding_gain() implemented without run_adding()\n");
    } else {
      log_comment("# run_adding() is not implemented\n");
    }
  }
  if (descriptor->deactivate) {
      log_comment("# deactivate() is needed\n");
  }
}

static void test (ladspa_plugin * plugin)
{
  plugin_code *code;
  const LADSPA_Descriptor *descriptor;

  log_comment("Testing %s:%ld\n", plugin->filename, plugin->index);

  code= load(plugin->filename, plugin->index);
  if (code == NULL) return;

  descriptor= code->descriptor;
  test_description(descriptor);
  test_lifecycle(descriptor);
  test_buffer(descriptor);
  test_controls(descriptor);
  test_control_extreme(descriptor);
  test_buffer_junk(descriptor);

  unload(code);
}

static void setup_buffers(LADSPA_Handle *handle,
                         const LADSPA_Descriptor *descriptor,
                         LADSPA_Data *buffers[], int size)
{
  unsigned long port;

  for (port= 0; port < descriptor->PortCount; ++port) {
    if (LADSPA_IS_PORT_AUDIO(descriptor->PortDescriptors[port])) {
      LADSPA_Data *buffer= calloc(size + 2, sizeof(LADSPA_Data));
      ((unsigned int *)buffer)[0]= DEADBEEF;
      ((unsigned int *)buffer)[size + 1]= DEADBEEF;

      if (LADSPA_IS_PORT_OUTPUT(descriptor->PortDescriptors[port])) {
        int k;
        for (k= 0; k<size; ++k)
          ((unsigned int *)buffer)[k+1]= FACEACHE;
      }

      descriptor->connect_port(handle, port, buffer + 1);
      buffers[port]= buffer;
    }
  }
}

static void check_buffers(const LADSPA_Descriptor *descriptor,
                          LADSPA_Data *buffers[],
                          int size, int do_input)
{
  unsigned long port;

  for (port= 0; port < descriptor->PortCount; ++port) {
    if (LADSPA_IS_PORT_AUDIO(descriptor->PortDescriptors[port])) {
      LADSPA_Data *buffer= buffers[port];

      if (((unsigned int *)buffer)[0] != DEADBEEF
       || ((unsigned int *)buffer)[size + 1] != DEADBEEF) {
        log_error("out of bounds memory write on port %lu\n", port);
      }

      if (LADSPA_IS_PORT_INPUT(descriptor->PortDescriptors[port])) {
        int k;

        for (k= 0; do_input && k<size; ++k) {
          if (buffer[k+1] != 0.0f) {
            log_error("illegal write to input port %lu\n", port);
            break;
          }
        }
      }
      if (LADSPA_IS_PORT_OUTPUT(descriptor->PortDescriptors[port])) {
        int k;

        for (k= 0; k<size; ++k) {
          if (((unsigned int *)buffer)[k+1] == FACEACHE) {
            log_error("unwritten output port %lu\n", port);
            break;
          }
        }
      }
      free(buffers[port]);
      buffers[port]= NULL;
    }
  }
}

static void setup_ports(LADSPA_Handle *handle,
                        const LADSPA_Descriptor *descriptor,
                        LADSPA_Data *controls)
{
  const LADSPA_PortRangeHint *hints = descriptor->PortRangeHints;
  unsigned long port;

  for (port= 0; port < descriptor->PortCount; ++port) {
    LADSPA_Data min, max, start;
    
    if (LADSPA_IS_PORT_AUDIO(descriptor->PortDescriptors[port]))
      continue; /* don't tamper with audio ports */

    if (LADSPA_IS_HINT_SAMPLE_RATE(hints[port].HintDescriptor)) {
      min= hints[port].LowerBound * (float) SRATE;
      max= hints[port].UpperBound * (float) SRATE;
    } else {
      min= hints[port].LowerBound;
      max= hints[port].UpperBound;
    }

    descriptor->connect_port(handle, port, &(controls[port]));
    if (LADSPA_IS_HINT_DEFAULT_MINIMUM(hints[port].HintDescriptor)) {
      start = min;
    } else if (LADSPA_IS_HINT_DEFAULT_LOW(hints[port].HintDescriptor)) {
      start = min * 0.75f + max * 0.25f;
    } else if (LADSPA_IS_HINT_DEFAULT_MIDDLE(hints[port].HintDescriptor)) {
      start = min * 0.5f + max * 0.5f;
    } else if (LADSPA_IS_HINT_DEFAULT_HIGH(hints[port].HintDescriptor)) {
      start = min * 0.25f + max * 0.75f;
    } else if (LADSPA_IS_HINT_DEFAULT_MAXIMUM(hints[port].HintDescriptor)) {
      start = max;
    } else if (LADSPA_IS_HINT_DEFAULT_0(hints[port].HintDescriptor)) {
      start = 0.0f;
    } else if (LADSPA_IS_HINT_DEFAULT_1(hints[port].HintDescriptor)) {
      start = 1.0f;
    } else if (LADSPA_IS_HINT_DEFAULT_100(hints[port].HintDescriptor)) {
      start = 100.0f;
    } else if (LADSPA_IS_HINT_DEFAULT_440(hints[port].HintDescriptor)) {
      start = 440.0f;
    } else if (LADSPA_IS_HINT_BOUNDED_BELOW(hints[port].HintDescriptor)
            && LADSPA_IS_HINT_BOUNDED_ABOVE(hints[port].HintDescriptor)) {
      start = (min + max) * 0.5f;
    } else {
      start = 0.0f;
    }
    controls[port]= start;
  }
}

static LADSPA_Handle instantiate(const LADSPA_Descriptor *descriptor,
                                 unsigned long rate)
{
  LADSPA_Handle handle;

  handle= descriptor->instantiate(descriptor, rate);
  if (handle == NULL) {
    log_error("Unexpected instantiation failure\n");
  }
  return handle;
}

static plugin_code * load (char *filename, long int index)
{
  LADSPA_Descriptor_Function descriptor_fn;
  plugin_code *code;

  code = g_new0(plugin_code, 1);

  code->library = dlopen(filename, RTLD_NOW);
  if (code->library == NULL) {
    log_error("dlopen: %s\n", dlerror());
    g_free(code);
    return NULL;
  }
  descriptor_fn = dlsym(code->library, "ladspa_descriptor");
  if (descriptor_fn == NULL) {
    log_error("dlsym: %s\n", dlerror());
    g_free(code);
    return NULL;
  }
  code->descriptor = descriptor_fn(index);

  return code;
}

static void unload (plugin_code *code) 
{
  dlclose(code->library);
  g_free(code);
}

static void find_all_plugins (void)
{
  char *ladspa_path, *directory;

  plugin_list = NULL; /* empty list */
  ladspa_path= getenv("LADSPA_PATH");
  if (ladspa_path != NULL) {
    ladspa_path = g_strdup(ladspa_path);

    directory = strtok(ladspa_path, ":");
    while (directory != NULL) {
      find_plugins(directory);
      directory = strtok(NULL, ":");
    }
    g_free(ladspa_path);
  }
}

static void find_plugins(char *path_entry)
{
  char lib_name[PATH_MAX];
  DIR *dir;
  struct dirent *dirent;
  
  dir= opendir(path_entry);
  if (dir == NULL) return;

  while ((dirent= readdir(dir))) {
    snprintf(lib_name, PATH_MAX, "%s/%s", path_entry, dirent->d_name);
    lib_plugins(lib_name);
  }

  closedir(dir);
  return;
}

static void lib_plugins(char *lib_name)
{
  LADSPA_Descriptor_Function descriptor_fn;
  const LADSPA_Descriptor *descriptor;
  void *library = NULL;
  ladspa_plugin *plugin;
  long int index;

  library = dlopen(lib_name, RTLD_LAZY);
  if (library == NULL) {
    return;
  }
  descriptor_fn = dlsym(library, "ladspa_descriptor");
  if (descriptor_fn == NULL) {
    dlclose(library);
    return;
  }

  for (index= 0; ; ++index) {
    descriptor= descriptor_fn(index);
    if (descriptor == NULL) {
      break;
    }
    plugin = g_new(ladspa_plugin, 1);
    plugin->filename= g_strdup(lib_name);
    plugin->index= index;
    plugin_list = g_slist_prepend(plugin_list, plugin);
  }
  dlclose(library);

}

void log_comment(const char *format, ...)
{
  va_list args;

  va_start(args, format);
  if (log_file) {
    vfprintf(log_file, format, args);
    fflush(log_file);
  }
  va_end(args);
  va_start(args, format);
  vfprintf(stdout, format, args);
  fflush(stdout);
  va_end(args);
}

void log_error(const char *format, ...)
{
  va_list args;

  va_start(args, format);
  if (log_file) {
    fputs("ERROR: ", log_file);
    vfprintf(log_file, format, args);
    fflush(log_file);
  }
  va_end(args);
  va_start(args, format);
  fputs("ERROR: ", stderr);
  vfprintf(stderr, format, args);
  fflush(stderr);
  va_end(args);
}

int main(int argc, char *argv[])
{
  GSList *list;
  int k;
  sigset_t signals;

  if (argc == 2 && !strcmp(argv[1], "--help")) {
    fprintf(stderr, "Usage: demolition [library.so]...\n"
            "Test one or more LADSPA plugins, either from LADSPA_PATH or\n"
            "from a list specified on the command line\n");
    return 0;
  } else if (argc > 1) {
    for (k= 1; k < argc; ++k) {
      lib_plugins(argv[k]);
    }
  } else {
    find_all_plugins();
  }

  log_file= fdopen(5, "a"); /* open special log descriptor if available */
  if (log_file == NULL) {
    fprintf(stderr, "No fd5, so no log written\n");
  }
  sigemptyset(&signals);
  sigaddset(&signals, SIGINT);
  sigprocmask(SIG_BLOCK, &signals, NULL);

  for (list= plugin_list; list != NULL; list = g_slist_next(list)) {
    pid_t pid;
    int status;
    ladspa_plugin *plugin = (ladspa_plugin *) list->data;

    pid = fork();
    if (pid) {
      waitpid(pid, &status, 0);
      if (WIFSIGNALED(status)) {
        switch (WTERMSIG(status)) {
          case SIGINT:
            log_error("User interrupted?\n");
            break;
          case SIGSEGV:
            log_error("Segfault\n");
            break;
          case SIGFPE:
            log_error("Floating point exception\n");
            break;
          case SIGALRM:
            log_error("Watchdog timeout\n");
            break;
          default:
            log_error("Unexpected signal %u\n", WTERMSIG(status));
            break;
        }
      }
    } else {
      sigprocmask(SIG_UNBLOCK, &signals, NULL);
      alarm(10);
      test(plugin);
      return 0;
    }
  }

  return 0;
}

